#!/usr/bin/env python

import sys
import syslogger

import nmi
import openswan

logger = syslogger.MultiLogger("NMVPNOpenswanProperties", 0)

def get_list_ipsec_connections(xnm):
    icr = openswan.IpsecConfReader('/etc/ipsec.conf')
    res = []
    if xnm:
        for section in icr:
            if section.type_name == 'conn' and section.has_key('x-nm') and section['x-nm'] == 'yes':
                res += [section.section_name]
    else:
        for section in icr:
            if section.type_name == 'conn' and \
               not (section.has_key('x-nm') and section['x-nm'] == 'no'):
                res += [section.section_name]
    return res
                
def get_connection_info(cname):
    icr = openswan.IpsecConfReader('/etc/ipsec.conf')
    conn = icr.sections[cname]
    ips = nmi.get_ipaddresses()
    ls = ""
    rs = ""
    if conn.has_key('leftsubnet'):
        ls = conn['leftsubnet']
    if conn.has_key('rightsubnet'):
        rs = conn['rightsubnet']            
    address = "%s // %s" % (conn['left'], conn['right'])
    subnet = "%s // %s" % (ls, rs)
    for la in nmi.get_ipaddresses():
        if la[0] == conn['left']:
            address = conn['right']
            subnet = rs
            break
        elif la[0] == conn['right']:
            address = conn['left']
            subnet = ls
            break
    return (address, mask)

        
if __name__ == '__main__':
    try:
        if sys.argv[1] == '--list-interfaces':
            print "\n".join(nmi.list_interfaces())
        elif sys.argv[1] == '--list-ipsec-connections':
            print "\n".join(get_list_ipsec_connections(False))
        elif sys.argv[1] == '--list-ipsec-connections-xnm':
            print "\n".join(get_list_ipsec_connections(True))
        elif sys.argv[1] == '--connection-info':
            name = sys.argv[2]
            print "\n".join(get_connection_info(name))
    except Exception, e:
        logger.error(str(e))
